// Copyright (C) 2020-2023 David Sugar, Tycho Softworks
// This code is licensed under MIT license

#include "drivers.hpp"
#include "memory.hpp"

#include <strings.hpp>

using namespace server;
using namespace std::literals::chrono_literals;

namespace {
const char *order[] = { "th",
    "1st", "2nd", "3rd", "4th", "5th",
    "6th", "7th", "8th", "9th", "10th",
    "11th", "12th", "13th", "14th", "15th",
    "16th", "17th", "18th", "19th",
    // extra entries for days of month...
    "20,th", "20,1st", "20,2nd", "20,3rd", "20,4th",
    "20,5th", "20,6th", "20,7th", "20,8th", "20,9th",
    "30,th", "30,1st"};

[[maybe_unused]] const char *weekdays[] = {"sunday", "monday", "tuesday", "wednasdays", "thursday", "friday", "saturday"};

auto low_order(char *out, size_t size, int num) {
    char buf[10];
    *out = 0;

    if(num >= 100) {
        buf[0] = '0' + (num / 100);
        buf[1] = 0;
        if(!str_append(out, size, buf, ",hundred,", nullptr))
            return false;
        num %= 100;
        if(!num)
            return str_append(out, size, "th,", nullptr);
    }

    if(num < 20)
        return str_append(out, size, order[num], ",", nullptr);

    buf[0] = '0' + num / 10;
    buf[1] = '0';
    buf[2] = ',';
    buf[3] = 0;
    return str_append(out, size, buf, order[num % 10], ",", nullptr);
}

auto low_number(char *out, size_t size, int num) {
    char buf[6];
    if(num >= 100) {
        buf[0] = '0' + (num / 100);
        buf[1] = 0;
        if(!str_append(out, size, buf, ",hundred,", nullptr))
            return false;
        num %= 100;
        if(!num)
            return true;
    }

    if(num < 10) {
        buf[0] = '0' + num;
        buf[1] = ',';
        buf[2] = 0;
    }
    else if(num < 20) {
        buf[0] = '1';
        buf[1] = '0' + num / 10;
        buf[2] = ',';
        buf[3] = 0;
    } else {
        buf[0] = '0' + num / 10;
        buf[1] = '0';
        buf[2] = ',';
        buf[3] = 0;
        if(num % 10) {
            buf[3] = '0' + (num % 10);
            buf[4] = ',';
            buf[5] = 0;
        }
    }
    return str_append(out, size, buf, nullptr);
}

auto speak_spell(char *out, size_t size, const char *v) {
    if(!v || strchr(v, ',') || mem_size(v) > (size / 2 - 2))
        return false;

    while(*v) {
        auto ch = tolower(*(v++));
        if(!isdigit(ch) && (ch < 'a' || ch > 'z'))
            continue;
        *(out++) = ch;
        *(out++) = ',';
    }
    *out = 0;
    return true;
}

auto speak_order(char *out, size_t size, const char *v) {
    if(!v || !*v || strchr(v, ',') || strchr(v, '.'))
        return false;

    auto num = atol(v);
    if(num < 0 || num > 999)
        return false;

    return low_order(out, size, num);
}

auto speak_number(char *out, size_t size, const char *v) {
    if(!v || !*v || strchr(v, ','))
        return false;

    *out = 0;
    auto zero = false;
    if(*v == '-') {
        str_append(out, size, "negative,", nullptr);
        ++v;
    }

    auto num = atol(v);
    if(num < 0)
        return false;

    if(num > 999999999) {
        zero = false;
        if(!low_number(out, size, num / 1000000000) || !str_append(out, size, "billion,", nullptr))
            return false;
        num %= 1000000000;
    }
    if(num > 999999) {
        zero = false;
        if(!low_number(out, size, num / 1000000) || !str_append(out, size, "million,", nullptr))
            return false;
        num %= 1000000;
    }
    if(num > 999) {
        zero = false;
        if(!low_number(out, size, num / 1000) || !str_append(out, size, "thousand,", nullptr))
            return false;

        num %= 1000;
    }

    if(num || zero)
        if(!low_number(out, size, num))
            return false;

    v = strchr(v, '.');
    if(!v || !*(++v))
        return true;
    if(!str_append(out, size, "point,", nullptr))
        return false;

    auto len = mem_size(out);
    return speak_spell(out + len, size - len, v);
}
} // end anon namespace

namespace server {
class en_plugin final : private Plugin {
public:
    en_plugin() : Plugin("en") {}

    auto speak(char *out, size_t size, const char *request, const char *value) -> bool final {
        if(eq(request, "&spell"))
            return speak_spell(out, size, value);
        else if(eq(request, "&number") || eq(request, "&decimal") || eq(request, "&count"))
            return speak_number(out, size, value);
        else if(eq(request, "&order"))
            return speak_order(out, size, value);
        else
            return false;
    }

    static en_plugin instance;
};
} // end namespace

NO_EXPORT en_plugin en_plugin::instance;
